import numpy as np
import json
from sklearn.metrics import confusion_matrix
import seaborn as sns
import matplotlib.pyplot as plt
import itertools
import pandas as pd
import os

from parameters import SAVE_MISCLASSIFIED_PATH, SAVE_CONFUSIONM_PATH, SAVE_PATH_LEARNING_CURVES, MODEL_NAME, AXIS_CONVERTION, EPOCHS, SPECIALIZATION

def get_learning_curves(train_losses, val_losses, train_accuracies, val_accuracies, savename = None):
    epochs = range(1, len(train_losses) + 1)

    plt.figure(figsize=(15,6))
    plt.subplot(1,2,1)
    plt.plot(epochs, train_losses, '-o', label='Training loss')
    plt.plot(epochs, val_losses, '-o', label='Validation loss')
    plt.legend()
    plt.title('Learning curves')
    plt.xlabel('Epoch')
    plt.ylabel('Loss')
    plt.xticks(epochs[::2], rotation=45)
    plt.subplot(1,2,2)
    plt.plot(epochs, train_accuracies, '-o', label='Training accuracy')
    plt.plot(epochs, val_accuracies, '-o', label='Validation accuracy')
    plt.legend()
    plt.title('Learning curves')
    plt.xlabel('Epoch')
    plt.ylabel('accuracy')
    plt.xticks(epochs[::2], rotation=45)
    if savename is not None:
        plt.savefig(SAVE_PATH_LEARNING_CURVES + savename + ".png", dpi=300, bbox_inches='tight')
    else:
        plt.savefig(SAVE_PATH_LEARNING_CURVES + ".png", dpi=300, bbox_inches='tight')


def get_misclassified_data(idx_pred, y_true, prob_pred, gt_labels_conf, test_image_paths, test_labels):
    error_indicator = idx_pred != y_true
    idxs = [i for i, x in enumerate(error_indicator) if x]
    misclassified_prob = list(itertools.compress(prob_pred, error_indicator))
    misclassified_gt_prob = list(itertools.compress(gt_labels_conf, error_indicator))
    misclassified_path = list(itertools.compress(test_image_paths, error_indicator))
    mislabelled = {}

    for i in range(0, len(idxs)):  # Loop through indexes directly
        true_label = test_labels[y_true[idxs[i]]]  # Get the true label once

        if isinstance(misclassified_path[i], str):
            image_path = misclassified_path[i].rsplit("_", 2)[0]       # Get the wrong path once
            axis = misclassified_path[i].rsplit("_", 2)[1] 
        else:
            image_path = misclassified_path[i][0].rsplit("_", 2)[0]
            axis = misclassified_path[i][0].rsplit("_", 2)[1] 

        model_name_entry = mislabelled.setdefault(true_label, {}).setdefault(image_path, {}).setdefault(AXIS_CONVERTION[axis], {})
        # Check and append `predicted_layer`
        if 'predicted_layer' in model_name_entry:
            if isinstance(model_name_entry['predicted_layer'], list):
                model_name_entry['predicted_layer'].append(test_labels[idx_pred[idxs[i]]])
            else:
                # Convert existing value to a list and append new value
                model_name_entry['predicted_layer'] = [model_name_entry['predicted_layer'], test_labels[idx_pred[idxs[i]]]]
        else:
            model_name_entry['predicted_layer'] = test_labels[idx_pred[idxs[i]]]

        # Check and append `confidence`
        if 'confidence' in model_name_entry:
            if isinstance(model_name_entry['confidence'], list):
                model_name_entry['confidence'].append(float(misclassified_prob[i]))
            else:
                model_name_entry['confidence'] = [model_name_entry['confidence'], float(misclassified_prob[i])]
        else:
            model_name_entry['confidence'] = float(misclassified_prob[i])

        # Check and append `gt_confidence`
        if 'gt_confidence' in model_name_entry:
            if isinstance(model_name_entry['gt_confidence'], list):
                model_name_entry['gt_confidence'].append(float(misclassified_gt_prob[i]))
            else:
                model_name_entry['gt_confidence'] = [model_name_entry['gt_confidence'], float(misclassified_gt_prob[i])]
        else:
            model_name_entry['gt_confidence'] = float(misclassified_gt_prob[i])

        # Save the mislabelled dictionary to a JSON file
        with open(SAVE_MISCLASSIFIED_PATH + ".json", 'w') as json_file:
            json.dump(mislabelled, json_file, indent=4)  # indent for pretty printing


def averaged_misclassified_data(final_predictions, ground_truth_dict, grouped_predictions, grouped_gt_predictions, test_labels):
    y_true = []
    y_pred_major = []
    misclassified_path = []
    misclassified_data = {}

    for base_name, predicted_class in final_predictions.items():
        y_true.append(ground_truth_dict[base_name]) 
        y_pred_major.append(predicted_class)

        if predicted_class != ground_truth_dict[base_name]:
            misclassified_path.append(base_name)
            misclassified_data[base_name] = {
                "gt": test_labels[int(ground_truth_dict[base_name])],  # Convert to standard Python int
                "each_pred": [test_labels[int(np.argmax(arr))] for arr in grouped_predictions[base_name]],  # Ensure each_pred is a list of Python ints
                "each_conf": [float(np.max(arr)) for arr in grouped_predictions[base_name]],  # Ensure each_conf is a list of Python floats
                "each_gt_conf": [float(np.max(arr)) for arr in grouped_gt_predictions[base_name]],  # Ensure each_conf is a list of Python floats
                "final_pred": int(final_predictions[base_name]),  # Convert to standard Python int
                "confidence": float(np.max(np.sum(np.array(grouped_predictions[base_name]), axis=0) / 3))  # Convert to Python float
            }

    y_true = np.array(y_true)
    y_pred_major = np.array(y_pred_major)

    # Save the mislabelled dictionary to a JSON file
    with open(SAVE_MISCLASSIFIED_PATH + "_averaging.json", 'w') as json_file:
        json.dump(misclassified_data, json_file, indent=4)  # indent for pretty printing

    return y_true, y_pred_major


def get_confusionM(y_true, y_pred_major, labels, test_set_len, averaging = False):
    num_errors = np.sum(y_true != y_pred_major)

    conf_matrix = confusion_matrix(y_true, y_pred_major)
    TP = conf_matrix.diagonal()
    P = conf_matrix.sum(axis=1)

    # Calculate balanced accuracy
    balanced_accuracy = sum(TP / P) / len(P) *100
    test_acc = 100-num_errors/test_set_len*100
    print(f'Test errors {num_errors} (out of {test_set_len:.0f})  {num_errors/test_set_len*100:0.2f}%')
    print(f'Test accuracy {test_acc:0.2f}%')
    print(f'Balanced accuracy {balanced_accuracy:0.2f}%')

    plt.figure(figsize=(8, 6))
    sns.heatmap(conf_matrix, annot=True, fmt='d', cmap='Blues',
                xticklabels=labels.values(),
                yticklabels=labels.values())
    plt.xlabel('Predicted Labels')
    plt.ylabel('True Labels')
    plt.title('Confusion Matrix')
    if averaging:
        plt.savefig(SAVE_CONFUSIONM_PATH + "averaging.png",  dpi=300, bbox_inches='tight')
    else:
        plt.savefig(SAVE_CONFUSIONM_PATH,  dpi=300, bbox_inches='tight')
    plt.show()

    return num_errors, test_acc, balanced_accuracy

def save_results(best_epoch, valid_loss, valid_acc, test_acc, balanced_test_acc, num_err, num_all_test, add_spec = "", file_name = "results.xlsx"):
    attributes = ["Model name", "Specialization", "Best epoch", "All epoch" ,"Validation loss", "Validation accuracy", 
                  "Test accuracy", "Balanced test accuracy", "Num of errors", "Errors out of"]
    new_row = {
        "Model name": MODEL_NAME,
        "Specialization": SPECIALIZATION + add_spec,
        "Best epoch": best_epoch,
        "All epoch": EPOCHS,  # Placeholder if you don't have this value yet
        "Validation loss": valid_loss,
        "Validation accuracy": valid_acc,
        "Test accuracy": test_acc,
        "Balanced test accuracy": balanced_test_acc,
        "Num of errors": num_err,  # Placeholder
        "Errors out of": num_all_test,  # Placeholder
    }

    # Check if the file exists
    if os.path.exists(file_name):
        # Read the existing table
        existing_df = pd.read_excel(file_name)
    else:
        # Create a new table if it doesn't exist
        existing_df = pd.DataFrame(columns=attributes)

    # Add the new row to the DataFrame
    existing_df = pd.concat([existing_df, pd.DataFrame([new_row])], ignore_index=True)

    # Save the updated DataFrame to the Excel file
    existing_df.to_excel(file_name, index=False)

    print(f"Table updated and saved to {file_name}")