import numpy as np
import random
from pprint import pprint
import os
import timm
from collections import Counter
import sys
import time

import torch
import torchvision.transforms as transforms
from torchvision.datasets import ImageFolder
from torch.utils.data import DataLoader
import torch.nn as nn
import torch.optim as optim

from parameters import (PIPELINE, TRAIN_FOLDER, VAL_FOLDER, TEST_FOLDER, INVIDUALS, DEVICE, BATCH_SIZE, WORKERS, MODEL_NAME, BEST_EPOCH, EPOCHS,
                        CHECKPOINTS_FOLDER, SAVE_PREDICTION_PATH, LOG_PATH)
from own_dataloader import (different_anatomical_plane_in_rgb, same_anatomical_plane_in_rgb, different_anatomical_plane_ensemble, 
                            ImageTripletDataset_to_one, same_anatomical_planes_in_RGB_ensemble, ImageTripletDataset_to_three, 
                            visualize_RGB_image_from_dataloader, get_data_distribution,visualize_ENSEMBLE_image_from_dataloader, RandomResample)
from train_functions import train_with_scaler
from test_functions import predict, averaging_prediction, predict_ENSEMBLE_with_confidence
from evaluation import get_learning_curves, averaged_misclassified_data, get_misclassified_data, get_confusionM, save_results

# FIX THE SEEDS
random.seed(42)
torch.manual_seed(42)
np.random.seed(42)

torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

if torch.cuda.is_available():
    torch.cuda.manual_seed_all(42)

os.environ['PYTHONHASHSEED'] = '42'

log_file = open(LOG_PATH, "w")
sys.stdout = log_file

# LOAD DATA
num_classes = get_data_distribution()

train_transform = transforms.Compose([
    RandomResample(scale_factor=2),
    transforms.ToTensor(),
    transforms.RandomHorizontalFlip(p=0.5),
    transforms.RandomVerticalFlip(p=0.5),
    transforms.RandomRotation(degrees=90, expand=True),

    transforms.Resize(224)
])
valid_transform = transforms.Compose([
    transforms.ToTensor(),
    transforms.Resize(224)
    
])
if PIPELINE == "basic":
  train_set = ImageFolder(TRAIN_FOLDER, transform = train_transform)
  val_set = ImageFolder(VAL_FOLDER, transform = valid_transform)
  test_set = ImageFolder(TEST_FOLDER, transform = valid_transform)
elif PIPELINE == "different_anatomical_plane_in_rgb":
  train_set = different_anatomical_plane_in_rgb(TRAIN_FOLDER, transform = train_transform)
  val_set = different_anatomical_plane_in_rgb(VAL_FOLDER, transform = valid_transform)
  test_set = different_anatomical_plane_in_rgb(TEST_FOLDER, transform = valid_transform)
elif PIPELINE == "same_anatomical_plane_in_rgb":
  train_set = same_anatomical_plane_in_rgb(TRAIN_FOLDER, transform = train_transform)
  val_set = same_anatomical_plane_in_rgb(VAL_FOLDER, transform = valid_transform)
  test_set = same_anatomical_plane_in_rgb(TEST_FOLDER, transform = valid_transform)
elif PIPELINE == "different_anatomical_plane_ensemble":
  train_set = different_anatomical_plane_ensemble(TRAIN_FOLDER, transform = train_transform)
  datasetAxial, datasetCoronial, datasetSagittal = different_anatomical_plane_ensemble.split_dataset_based_on_number(train_set.samples)
  train_datasetAxial = ImageTripletDataset_to_one(datasetAxial, transform = train_transform)
  train_datasetCoronial = ImageTripletDataset_to_one(datasetCoronial, transform = train_transform)
  train_datasetSagittal = ImageTripletDataset_to_one(datasetSagittal, transform = train_transform)

  val_set = different_anatomical_plane_ensemble(VAL_FOLDER, transform = valid_transform)
  datasetAxial, datasetCoronial, datasetSagittal = different_anatomical_plane_ensemble.split_dataset_based_on_number(val_set.samples)
  val_datasetAxial = ImageTripletDataset_to_one(datasetAxial, transform = valid_transform)
  val_datasetCoronial = ImageTripletDataset_to_one(datasetCoronial, transform = valid_transform)
  val_datasetSagittal = ImageTripletDataset_to_one(datasetSagittal, transform = valid_transform)

  test_set = different_anatomical_plane_ensemble(TEST_FOLDER, transform = valid_transform)
  datasetAxial, datasetCoronial, datasetSagittal = different_anatomical_plane_ensemble.split_dataset_based_on_number(test_set.samples)
  test_datasetAxial = ImageTripletDataset_to_one(datasetAxial, transform = valid_transform)
  test_datasetCoronial = ImageTripletDataset_to_one(datasetCoronial, transform = valid_transform)
  test_datasetSagittal = ImageTripletDataset_to_one(datasetSagittal, transform = valid_transform)
elif PIPELINE == "ensemble_same_anatomical_planes_in_RGB":
  train_set = same_anatomical_planes_in_RGB_ensemble(TRAIN_FOLDER, transform = train_transform)
  datasetAxial, datasetCoronial, datasetSagittal = same_anatomical_planes_in_RGB_ensemble.split_dataset_based_on_number(train_set.samples)
  train_datasetAxial = ImageTripletDataset_to_three(datasetAxial, transform = train_transform)
  train_datasetCoronial = ImageTripletDataset_to_three(datasetCoronial, transform = train_transform)
  train_datasetSagittal = ImageTripletDataset_to_three(datasetSagittal, transform = train_transform)

  val_set = same_anatomical_planes_in_RGB_ensemble(VAL_FOLDER, transform = valid_transform)
  datasetAxial, datasetCoronial, datasetSagittal = same_anatomical_planes_in_RGB_ensemble.split_dataset_based_on_number(val_set.samples)
  val_datasetAxial = ImageTripletDataset_to_three(datasetAxial, transform = valid_transform)
  val_datasetCoronial = ImageTripletDataset_to_three(datasetCoronial, transform = valid_transform)
  val_datasetSagittal = ImageTripletDataset_to_three(datasetSagittal, transform = valid_transform)

  test_set = same_anatomical_planes_in_RGB_ensemble(TEST_FOLDER, transform = valid_transform)
  datasetAxial, datasetCoronial, datasetSagittal = same_anatomical_planes_in_RGB_ensemble.split_dataset_based_on_number(test_set.samples)
  test_datasetAxial = ImageTripletDataset_to_three(datasetAxial, transform = valid_transform)
  test_datasetCoronial = ImageTripletDataset_to_three(datasetCoronial, transform = valid_transform)
  test_datasetSagittal = ImageTripletDataset_to_three(datasetSagittal, transform = valid_transform)

if PIPELINE == "basic" or PIPELINE == "different_anatomical_plane_in_rgb" or PIPELINE == "same_anatomical_plane_in_rgb":
  train_loader = DataLoader(train_set, batch_size = BATCH_SIZE, shuffle = True, num_workers=WORKERS, drop_last=True)
  val_loader = DataLoader(val_set, batch_size = BATCH_SIZE,  shuffle = False, num_workers=WORKERS)
  test_loader = DataLoader(test_set, batch_size = 1,  shuffle = False, num_workers=WORKERS)

  labels = [label for _, label in train_set.samples]
  test_image_paths = [path for path, _ in test_set.samples]
  print(f'test samples: {len(test_set)}')

  test_labels = {v: k for k, v in test_set.class_to_idx.items()}
  print(f"Len of test labels: {len(test_labels)}, labels: {test_labels}")

  visualize_RGB_image_from_dataloader(train_loader)

elif PIPELINE == "different_anatomical_plane_ensemble" or PIPELINE == "ensemble_same_anatomical_planes_in_RGB":
  train_axial_loader   =  DataLoader(train_datasetAxial, batch_size = BATCH_SIZE, shuffle = True, num_workers=WORKERS, drop_last=True)
  train_coronial_loader = DataLoader(train_datasetCoronial, batch_size = BATCH_SIZE, shuffle = True, num_workers=WORKERS, drop_last=True)
  train_sagittal_loader = DataLoader(train_datasetSagittal, batch_size = BATCH_SIZE, shuffle = True, num_workers=WORKERS, drop_last=True)

  val_axial_loader   =  DataLoader(val_datasetAxial, batch_size = BATCH_SIZE, shuffle = False, num_workers=WORKERS)
  val_coronial_loader = DataLoader(val_datasetCoronial, batch_size = BATCH_SIZE, shuffle = False, num_workers=WORKERS)
  val_sagittal_loader = DataLoader(val_datasetSagittal, batch_size = BATCH_SIZE, shuffle = False, num_workers=WORKERS)

  test_axial_loader   =  DataLoader(test_datasetAxial, batch_size = 1, shuffle = False, num_workers=WORKERS)
  test_coronial_loader = DataLoader(test_datasetCoronial, batch_size = 1, shuffle = False, num_workers=WORKERS)
  test_sagittal_loader = DataLoader(test_datasetSagittal, batch_size = 1, shuffle = False, num_workers=WORKERS)

  labels = [label for _, label in train_datasetAxial.samples]
  test_image_paths = [path for path, _ in test_datasetAxial.samples]
  print(f'test samples: {len(test_datasetAxial)}')
  test_labels = {v: k for k, v in test_datasetAxial.class_to_idx.items()}
  print(f"Len of test labels: {len(test_labels)}, labels: {test_labels}")
  if PIPELINE == "different_anatomical_plane_ensemble":
    visualize_ENSEMBLE_image_from_dataloader(train_axial_loader, train_coronial_loader, train_sagittal_loader)
  else:
    visualize_RGB_image_from_dataloader(train_axial_loader, "_model1")
    visualize_RGB_image_from_dataloader(train_coronial_loader, "_model2")
    visualize_RGB_image_from_dataloader(train_sagittal_loader, "_model3")

num_in_class_dict = dict(Counter(labels))
num_in_class = np.zeros([1,len(num_in_class_dict)])
for i in range(0, len(num_in_class_dict)):
  num_in_class[0, i] = num_in_class_dict[i]

class_weights = 1-(num_in_class/num_in_class.sum()).squeeze()
class_weights_tensor = torch.Tensor(class_weights).to(DEVICE)

print("Class distribution (number of samples per class):", num_in_class_dict)
print("Class distribution as numpy array:", num_in_class)
print("Class weights:", class_weights_tensor)

torch.cuda.is_available()

# TRAINING
start_time = time.time()
if PIPELINE == "basic" or PIPELINE == "different_anatomical_plane_in_rgb" or PIPELINE == "same_anatomical_plane_in_rgb":
  model = timm.create_model(MODEL_NAME, pretrained=True, num_classes=num_classes)
  model.to(DEVICE)
  criterion_balanced = nn.CrossEntropyLoss(weight = class_weights_tensor)
  optimizer_Adam = optim.Adam(model.parameters(), 1e-3)
  scaler = torch.amp.GradScaler(DEVICE)

  if not os.path.exists(CHECKPOINTS_FOLDER):
      os.mkdir(CHECKPOINTS_FOLDER)

  train_losses, val_losses, train_accuracies, val_accuracies, BEST_EPOCH = train_with_scaler(model, train_loader, val_loader, optimizer_Adam, criterion_balanced, 
                                                                                            EPOCHS, scaler, DEVICE, checkpoints_foler=CHECKPOINTS_FOLDER)

  get_learning_curves(train_losses, val_losses, train_accuracies, val_accuracies)
  print(f"Best epoch: {BEST_EPOCH}")
  print(f"Best val_losses: {val_losses[BEST_EPOCH-1]}, Best val_accuracies: {val_accuracies[BEST_EPOCH-1]}")

elif PIPELINE == "different_anatomical_plane_ensemble" or PIPELINE == "ensemble_same_anatomical_planes_in_RGB":
  model1 = timm.create_model(MODEL_NAME, pretrained=True, num_classes=num_classes)
  model1.to(DEVICE)
  criterion_balanced1 = nn.CrossEntropyLoss(weight = class_weights_tensor)
  optimizer_Adam1 = optim.Adam(model1.parameters(), 1e-3)
  scaler1 = torch.amp.GradScaler('cuda')
  checkpoints_folder1 = CHECKPOINTS_FOLDER + '_ensemble_1_checkpoints'
  if not os.path.exists(checkpoints_folder1):
      os.mkdir(checkpoints_folder1)


  train_losses1, val_losses1, train_accuracies1, val_accuracies1, BEST_EPOCH1 = train_with_scaler(model1, train_axial_loader, val_axial_loader,
                                                                                              optimizer_Adam1, criterion_balanced1, EPOCHS, scaler1,
                                                                                              DEVICE, checkpoints_foler=checkpoints_folder1)

  model2 = timm.create_model(MODEL_NAME, pretrained=True, num_classes=num_classes)
  model2.to(DEVICE)
  criterion_balanced2 = nn.CrossEntropyLoss(weight = class_weights_tensor)
  optimizer_Adam2 = optim.Adam(model2.parameters(), 1e-3)
  scaler2 = torch.amp.GradScaler('cuda')
  checkpoints_folder2 = CHECKPOINTS_FOLDER + '_ensemble_2_checkpoints'

  if not os.path.exists(checkpoints_folder2):
      os.mkdir(checkpoints_folder2)
  train_losses2, val_losses2, train_accuracies2, val_accuracies2, BEST_EPOCH2 = train_with_scaler(model2, train_coronial_loader, val_coronial_loader,
                                                                                              optimizer_Adam2, criterion_balanced2, EPOCHS, scaler2,
                                                                                              DEVICE, checkpoints_foler=checkpoints_folder2)

  model3 = timm.create_model(MODEL_NAME, pretrained=True, num_classes=num_classes)
  model3.to(DEVICE)
  criterion_balanced3 = nn.CrossEntropyLoss(weight = class_weights_tensor)
  optimizer_Adam3 = optim.Adam(model3.parameters(), 1e-3)
  scaler3 = torch.amp.GradScaler('cuda')
  checkpoints_folder3 = CHECKPOINTS_FOLDER + 'ensemble_3_checkpoints'

  if not os.path.exists(checkpoints_folder3):
      os.mkdir(checkpoints_folder3)
  train_losses3, val_losses3, train_accuracies3, val_accuracies3, BEST_EPOCH3 = train_with_scaler(model3, train_sagittal_loader, val_sagittal_loader,
                                                                                            optimizer_Adam3, criterion_balanced3, EPOCHS, scaler3,
                                                                                            DEVICE, checkpoints_foler=checkpoints_folder3)
  
  get_learning_curves(train_losses1, val_losses1, train_accuracies1, val_accuracies1, "_1_model")
  get_learning_curves(train_losses2, val_losses2, train_accuracies2, val_accuracies2, "_2_model")
  get_learning_curves(train_losses3, val_losses3, train_accuracies3, val_accuracies3, "_3_model")
  print(f"Best epoch_model1: {BEST_EPOCH1}, Best epoch_model2: {BEST_EPOCH2}, Best epoch_model3: {BEST_EPOCH3}")
  print(f"MODEL1: Best val_losses: {val_losses1[BEST_EPOCH1-1]}, Best val_accuracies: {val_accuracies1[BEST_EPOCH1-1]}")
  print(f"MODEL2: Best val_losses: {val_losses2[BEST_EPOCH2-1]}, Best val_accuracies: {val_accuracies2[BEST_EPOCH2-1]}")
  print(f"MODEL3: Best val_losses: {val_losses3[BEST_EPOCH3-1]}, Best val_accuracies: {val_accuracies3[BEST_EPOCH3-1]}")

print(f"Training time: {time.time() - start_time} s")

# TESTING
if PIPELINE == "basic" or PIPELINE == "different_anatomical_plane_in_rgb" or PIPELINE == "same_anatomical_plane_in_rgb":
  model = torch.load(CHECKPOINTS_FOLDER+f'/avp_{BEST_EPOCH:03d}.pkl')
  model.to(DEVICE)

  print("prediction started")
  y_pred, gt_labels_conf = predict(model, test_loader)
  np.save(SAVE_PREDICTION_PATH + str(BEST_EPOCH), y_pred)
  print("prediction saved")

elif PIPELINE == "different_anatomical_plane_ensemble" or PIPELINE == "ensemble_same_anatomical_planes_in_RGB":
  model1 = torch.load(checkpoints_folder1+f'/avp_{BEST_EPOCH1:03d}.pkl')
  model2 = torch.load(checkpoints_folder2+f'/avp_{BEST_EPOCH2:03d}.pkl')
  model3 = torch.load(checkpoints_folder3+f'/avp_{BEST_EPOCH3:03d}.pkl')
  y_pred_majority, y_true, confidence_scores_majority, gt_labels_conf = predict_ENSEMBLE_with_confidence(model1, model2, model3, test_axial_loader, test_coronial_loader, test_sagittal_loader, DEVICE)

#val_losses = None
#val_accuracies = None
#EVALUATION
if PIPELINE == "different_anatomical_plane_in_rgb":
  # different and same anatomical planes in RGB
  y_pred_majority = y_pred.argmax(axis=1) # find the argmax of each of the predictions
  confidence_scores_majority = y_pred.max(axis = 1) #get the confidence score
  y_true = [label for _, label in test_set.samples] # get the true labels and convert to numpy
  num_errors, test_acc, balanced_accuracy = get_confusionM(y_true, y_pred_majority, test_labels, len(test_set))
  get_misclassified_data(y_pred_majority, y_true, confidence_scores_majority, gt_labels_conf, test_image_paths, test_labels)
  save_results(BEST_EPOCH, val_losses, val_accuracies, test_acc, balanced_accuracy, num_errors, len(test_set))

elif PIPELINE == "basic" or PIPELINE == "same_anatomical_plane_in_rgb":
  y_pred_majority = y_pred.argmax(axis=1) # find the argmax of each of the predictions
  confidence_scores_majority = y_pred.max(axis = 1) #get the confidence score
  y_true = [label for _, label in test_set.samples] # get the true labels and convert to numpy
  num_errors, test_acc, balanced_accuracy = get_confusionM(y_true, y_pred_majority, test_labels, len(test_set))
  get_misclassified_data(y_pred_majority, y_true, confidence_scores_majority, gt_labels_conf, test_image_paths, test_labels)
  save_results(BEST_EPOCH, val_losses, val_accuracies, test_acc, balanced_accuracy, num_errors, len(test_set))

  # averaging evaluation -> simaknal meg a same anatomical planes in RGB
  grouped_predictions, ground_truth_dict, final_predictions, grouped_gt_predictions = averaging_prediction(y_pred, test_set, test_image_paths)
  y_true, y_pred_major = averaged_misclassified_data(final_predictions, ground_truth_dict, grouped_predictions, grouped_gt_predictions, test_labels)
  AVG_num_errors, AVG_test_acc, AVG_balanced_accuracy = get_confusionM(y_true, y_pred_major, test_labels, INVIDUALS, averaging=True)
  save_results(BEST_EPOCH, val_losses, val_accuracies, AVG_test_acc, AVG_balanced_accuracy, AVG_num_errors, INVIDUALS)
  
elif PIPELINE == "different_anatomical_plane_ensemble" or PIPELINE == "ensemble_same_anatomical_planes_in_RGB":
  # different and same anatomical plane ensemble
  num_errors, test_acc, balanced_accuracy = get_confusionM(y_true, y_pred_majority, test_labels, len(test_datasetAxial))
  get_misclassified_data(y_pred_majority, y_true, confidence_scores_majority, gt_labels_conf, test_image_paths, test_labels)
  save_results(BEST_EPOCH1, val_losses1, val_accuracies1, test_acc, balanced_accuracy, num_errors, len(test_datasetAxial))
  save_results(BEST_EPOCH2, val_losses2, val_accuracies2, test_acc, balanced_accuracy, num_errors, len(test_datasetCoronial))
  save_results(BEST_EPOCH3, val_losses3, val_accuracies3, test_acc, balanced_accuracy, num_errors, len(test_datasetSagittal))

sys.stdout = sys.__stdout__  # Reset stdout back to normal
log_file.close()

print("DONE")