from torch.utils.data import Dataset
import numpy as np
import os
from PIL import Image
import matplotlib.pyplot as plt
import random

from parameters import SAVE_INPUT_VISUALIZATION_PATH, TRAIN_FOLDER, VAL_FOLDER, TEST_FOLDER

# define custom resample class to change image resolution without rescaling
class RandomResample:
    def __init__(self, scale_factor):
        self.scale_factor = random.uniform(0,scale_factor)
    
    def __call__(self, img):
        # Downsample
        if isinstance(img, np.ndarray):
            img = Image.fromarray(img)
        
        width, height = img.size
        downscaled_size = (int(width / self.scale_factor), int(height / self.scale_factor))
        
        # Downsample the image
        img_downsampled = img.resize(downscaled_size)
        
        # Upsample back to the original size
        img_upsampled = img_downsampled.resize((width, height))
        
        return img_upsampled
    
def get_data_distribution():
    train_categories = os.listdir(TRAIN_FOLDER)
    val_categories = os.listdir(VAL_FOLDER)
    test_categories = os.listdir(TEST_FOLDER)

    print("Train image distribution: ")
    class_num_in_train = []
    for i in range(0, len(train_categories)):
        CLASS_FOLDER = TRAIN_FOLDER + '/' + train_categories[i]
        class_elements = os.listdir(CLASS_FOLDER)
        class_num_in_train.append(len(class_elements))
        print(f' {train_categories[i]}: {class_num_in_train[i]}')
    
    print("Validation image distribution: ")
    class_num_in_val = []
    for i in range(0, len(val_categories)):
        CLASS_FOLDER = VAL_FOLDER + '/' + val_categories[i]
        class_elements = os.listdir(CLASS_FOLDER)
        class_num_in_val.append(len(class_elements))
        print(f' {val_categories[i]}: {class_num_in_val[i]}')
    
    print("Test image distribution: ")
    class_num_in_test = []
    for i in range(0, len(test_categories)):
        CLASS_FOLDER = TEST_FOLDER + '/' + test_categories[i]
        class_elements = os.listdir(CLASS_FOLDER)
        class_num_in_test.append(len(class_elements))
        print(f' {test_categories[i]}: {class_num_in_test[i]}')

    return len(class_num_in_train)

def visualize_RGB_image_from_dataloader(dataloader, spec_name = None):
    # Get a batch from the dataloader
    images, labels = next(iter(dataloader))
    # Take the first image from the batch (assuming the batch size > 0)
    img = images[0].numpy()  # Convert the tensor to a numpy array
    
    # Rearrange the image shape from (C, H, W) to (H, W, C) for plotting
    img_rgb = np.transpose(img, (1, 2, 0))

    # Plot the RGB image and individual channels
    fig, axes = plt.subplots(1, 4, figsize=(20, 5))

    # Plot the RGB image
    axes[0].imshow(img_rgb)
    axes[0].set_title(f"RGB Image: {labels[0]}")
    axes[0].axis('off')  # Hide axes for clarity

    # Plot the Red channel
    axes[1].imshow(img[0], cmap='Reds')  # First channel (Red)
    axes[1].set_title('Axial - Red Channel')
    axes[1].axis('off')

    # Plot the Green channel
    axes[2].imshow(img[1], cmap='Greens')  # Second channel (Green)
    axes[2].set_title('Coronal - Green Channel')
    axes[2].axis('off')

    # Plot the Blue channel
    axes[3].imshow(img[2], cmap='Blues')  # Third channel (Blue)
    axes[3].set_title('Sagittal - Blue Channel')
    axes[3].axis('off')

    plt.tight_layout()
    if spec_name is not None:
        plt.savefig(SAVE_INPUT_VISUALIZATION_PATH + spec_name + ".png",  dpi=300, bbox_inches='tight')
    else:
        plt.savefig(SAVE_INPUT_VISUALIZATION_PATH + ".png",  dpi=300, bbox_inches='tight')
    plt.show()

def visualize_ENSEMBLE_image_from_dataloader(dataloader1, dataloader2, dataloader3):
    # Get a batch from the dataloader
    images, labels = next(iter(dataloader1))
    # Take the first image from the batch (assuming the batch size > 0)
    img = images[0].numpy()  # Convert the tensor to a numpy array
    img_transposed = img.transpose(1, 2, 0)
    img_pil = Image.fromarray((img_transposed * 255).astype(np.uint8))
    gray_img_pil = img_pil.convert("L")
    img = np.array(gray_img_pil)
    # Rearrange the image shape from (C, H, W) to (H, W, C) for plotting
    #img_rgb = np.transpose(img, (1, 2, 0))

    images2, _ = next(iter(dataloader2))
    img2 = images2[0].numpy()  # Convert the tensor to a numpy array
    img2_transposed = img2.transpose(1, 2, 0)
    img2_pil = Image.fromarray((img2_transposed * 255).astype(np.uint8))
    gray_img2_pil = img2_pil.convert("L")
    img2 = np.array(gray_img2_pil)

    images3, _ = next(iter(dataloader3))
    # Take the first image from the batch (assuming the batch size > 0)
    img3 = images3[0].numpy()  # Convert the tensor to a numpy array
    img3_transposed = img3.transpose(1, 2, 0)
    img3_pil = Image.fromarray((img3_transposed * 255).astype(np.uint8))
    gray_img3_pil = img3_pil.convert("L")
    img3 = np.array(gray_img3_pil)
    
    rgb_image = np.stack((img, img2, img3), axis=-1)

    # Plot the RGB image and individual channels
    fig, axes = plt.subplots(1, 4, figsize=(20, 5))

    # Plot the RGB image
    axes[0].imshow(rgb_image.squeeze())
    axes[0].set_title(f"RGB Image: {labels[0]}")
    axes[0].axis('off')  # Hide axes for clarity

    # Plot the Red channel
    axes[1].imshow(img.squeeze(), cmap='Reds')  # First channel (Red)
    axes[1].set_title('Axial - Red Channel')
    axes[1].axis('off')

    # Plot the Green channel
    axes[2].imshow(img2.squeeze(), cmap='Greens')  # Second channel (Green)
    axes[2].set_title('Coronal - Green Channel')
    axes[2].axis('off')

    # Plot the Blue channel
    axes[3].imshow(img3.squeeze(), cmap='Blues')  # Third channel (Blue)
    axes[3].set_title('Sagittal - Blue Channel')
    axes[3].axis('off')

    plt.tight_layout()
    plt.savefig(SAVE_INPUT_VISUALIZATION_PATH + ".png",  dpi=300, bbox_inches='tight')
    plt.show()

class different_anatomical_plane_in_rgb(Dataset):
    def __init__(self, root_dir, transform=None):
        """
        Args:
            root_dir (str): Directory with all the images and subfolders (A, B, C, D, E).
            transform (callable, optional): Optional transform to be applied on a sample.
        """
        self.root_dir = root_dir
        self.transform = transform
        self.class_to_idx = self._create_class_to_idx()
        self.samples = self._load_samples()

    def _create_class_to_idx(self):
        """Creates a mapping from folder names to integer labels."""
        class_to_idx = {
            'FLAIR': 0,
            'FLAIRCE': 1,
            'OTHER': 2,
            'T1w': 3,
            'T1wCE': 4,
            'T2star': 5,
            'T2w': 6
        }
        return class_to_idx

    def _load_samples(self):
        """Collects all the image paths from the subfolders and associates labels based on folder name."""
        samples = []
        
        # List all the subfolders A, B, C, D, E, etc.
        for folder in os.listdir(self.root_dir):
            folder_path = os.path.join(self.root_dir, folder)
            
            if os.path.isdir(folder_path):  # Check if it is a folder
                # List all files in the folder and sort them
                folder_items =  os.listdir(folder_path)
                main_file_names = set([path.rsplit("_", 2)[0] for path in folder_items])
                for name in main_file_names:
                    if name != "index.html":
                        files_to_connect = sorted([path for path in folder_items if name in path])
                        img_paths = [os.path.join(folder_path, files_to_connect[0]),
                                    os.path.join(folder_path, files_to_connect[1]),
                                    os.path.join(folder_path, files_to_connect[2])]

                        label = self.class_to_idx[folder]
                        samples.append((img_paths, label))
        
        return samples

    def __len__(self):
        """Returns the number of samples."""
        return len(self.samples)

    def __getitem__(self, idx):
        """Fetches an item and applies any transformations."""
        # Get the paths for the three images and label
        img_paths, label = self.samples[idx]
        
        # Load the three grayscale images
        img0 = Image.open(img_paths[0]).convert('L')  # Convert to grayscale
        img1 = Image.open(img_paths[1]).convert('L')
        img2 = Image.open(img_paths[2]).convert('L')
        
        # Convert to numpy arrays
        img0_array = np.array(img0)
        img1_array = np.array(img1)
        img2_array = np.array(img2)
        
        # Stack into an RGB image
        rgb_image = np.stack((img0_array, img1_array, img2_array), axis=-1)
        
        # Convert numpy array to PyTorch tensor (C, H, W) format
        #rgb_image = torch.tensor(rgb_image).permute(2, 0, 1).float()  # Convert to C, H, W
        if self.transform:
            rgb_image = self.transform(rgb_image)
        
        return rgb_image, label

class different_anatomical_plane_ensemble(Dataset):
    def __init__(self, root_dir, transform=None):
        """
        Args:
            root_dir (str): Directory with all the images and subfolders (A, B, C, D, E).
            transform (callable, optional): Optional transform to be applied on a sample.
        """
        self.root_dir = root_dir
        self.transform = transform
        self.class_to_idx = self._create_class_to_idx()
        self.samples = self._load_samples()

    def _create_class_to_idx(self):
        """Creates a mapping from folder names to integer labels."""
        class_to_idx = {
            'FLAIR': 0,
            'FLAIRCE': 1,
            'OTHER': 2,
            'T1w': 3,
            'T1wCE': 4,
            'T2star': 5,
            'T2w': 6
        }
        return class_to_idx

    def _load_samples(self):
        """Collects all the image paths from the subfolders and associates labels based on folder name."""
        samples = []
        
        # List all the subfolders A, B, C, D, E, etc.
        for folder in os.listdir(self.root_dir):
            folder_path = os.path.join(self.root_dir, folder)
            
            if os.path.isdir(folder_path):  # Check if it is a folder
                # List all files in the folder and sort them
                folder_items =  os.listdir(folder_path)
                main_file_names = set([path.rsplit("_", 2)[0] for path in folder_items])
                for name in main_file_names:
                    if name != "index.html":
                        files_to_connect = sorted([path for path in folder_items if name in path])
                        for i in range(0, len(files_to_connect)):
                            img_paths = [os.path.join(folder_path, files_to_connect[i])]
                            group = i
                            label = self.class_to_idx[folder]
                            samples.append((img_paths, label, group))
        
        return samples

    def __len__(self):
        """Returns the number of samples."""
        return len(self.samples)

    def __getitem__(self, idx):
        """Fetches an item and applies any transformations."""
        # Get the paths for the three images and label
        img_paths, label, _ = self.samples[idx]
        
        # Load the three grayscale images
        img0 = Image.open(img_paths[0]).convert('L')  # Convert to grayscale
        img1 = Image.open(img_paths[1]).convert('L')
        img2 = Image.open(img_paths[2]).convert('L')
        
        # Convert to numpy arrays
        img0_array = np.array(img0)
        img1_array = np.array(img1)
        img2_array = np.array(img2)
        
        # Stack into an RGB image
        rgb_image = np.stack((img0_array, img1_array, img2_array), axis=-1)
        
        # Convert numpy array to PyTorch tensor (C, H, W) format
        #rgb_image = torch.tensor(rgb_image).permute(2, 0, 1).float()  # Convert to C, H, W
        if self.transform:
            rgb_image = self.transform(rgb_image)
        
        return rgb_image, label
    
    def split_dataset_based_on_number(samples):
        """Splits the dataset into 3 datasets based on the number in the filename (0, 1, 2)."""
        datasetAxial = []
        datasetCoronial = []
        datasetSagittal = []
        
        # Divide samples into 3 datasets based on the `numberY` (0, 1, 2)
        for img_paths, label, group in samples:
            if group == 0:
                datasetAxial.append((img_paths, label))
            elif group == 1:
                datasetCoronial.append((img_paths, label))
            elif group == 2:
                datasetSagittal.append((img_paths, label))  # 2nd group for datasetC
        
        return datasetAxial, datasetCoronial, datasetSagittal

class ImageTripletDataset_to_one(Dataset):
    def __init__(self, samples, transform=None):
        """
        Args:
            samples (list): List of tuples where each tuple is (img_paths, label).
            transform (callable, optional): Optional transform to be applied on a sample.
        """
        self.class_to_idx = self._create_class_to_idx()
        self.samples = samples
        self.transform = transform

    def _create_class_to_idx(self):
        """Creates a mapping from folder names to integer labels."""
        class_to_idx = {
            'FLAIR': 0,
            'FLAIRCE': 1,
            'OTHER': 2,
            'T1w': 3,
            'T1wCE': 4,
            'T2star': 5,
            'T2w': 6
        }
        return class_to_idx
    
    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        # Get the paths for the three images and label
        img_paths, label = self.samples[idx]
        
        # Load the three grayscale images
        img0 = Image.open(img_paths[0]).convert('L')  # Convert to grayscale
        rgb_image = img0.convert('RGB')
        # Convert to numpy arrays
        img0_array = np.array(rgb_image)
        
        # Apply any transformations if specified
        if self.transform:
            img0_array = self.transform(img0_array)

        return img0_array, label
    
class same_anatomical_planes_in_RGB_ensemble(Dataset):
    def __init__(self, root_dir, transform=None):
        """
        Args:
            root_dir (str): Directory with all the images and subfolders (A, B, C, D, E).
            transform (callable, optional): Optional transform to be applied on a sample.
        """
        self.root_dir = root_dir
        self.transform = transform
        self.class_to_idx = self._create_class_to_idx()
        self.samples = self._load_samples()

    def _create_class_to_idx(self):
        """Creates a mapping from folder names to integer labels."""
        class_to_idx = {
            'FLAIR': 0,
            'FLAIRCE': 1,
            'OTHER': 2,
            'T1w': 3,
            'T1wCE': 4,
            'T2star': 5,
            'T2w': 6
        }
        return class_to_idx

    def _load_samples(self):
        """Collects all the image paths from the subfolders and associates labels based on folder name."""
        samples = []
        
        # List all the subfolders A, B, C, D, E, etc.
        for folder in os.listdir(self.root_dir):
            folder_path = os.path.join(self.root_dir, folder)
            
            if os.path.isdir(folder_path):  # Check if it is a folder
                # List all files in the folder and sort them
                folder_items =  os.listdir(folder_path)
                main_file_names = set([path.rsplit("_", 2)[0] for path in folder_items])
                for name in main_file_names:
                    if name != "index.html":
                        files_to_connect = sorted([path for path in folder_items if name in path])
                        for i in range(0, len(files_to_connect), 3):
                            img_paths = [os.path.join(folder_path, files_to_connect[i]),
                                        os.path.join(folder_path, files_to_connect[i+1]),
                                        os.path.join(folder_path, files_to_connect[i+2])]
                            group = int(i/3)
                            label = self.class_to_idx[folder]
                            samples.append((img_paths, label, group))
        
        return samples

    def __len__(self):
        """Returns the number of samples."""
        return len(self.samples)

    def __getitem__(self, idx):
        """Fetches an item and applies any transformations."""
        # Get the paths for the three images and label
        img_paths, label, _ = self.samples[idx]
        
        # Load the three grayscale images
        img0 = Image.open(img_paths[0]).convert('L')  # Convert to grayscale
        img1 = Image.open(img_paths[1]).convert('L')
        img2 = Image.open(img_paths[2]).convert('L')
        
        # Convert to numpy arrays
        img0_array = np.array(img0)
        img1_array = np.array(img1)
        img2_array = np.array(img2)
        
        # Stack into an RGB image
        rgb_image = np.stack((img0_array, img1_array, img2_array), axis=-1)
        
        # Convert numpy array to PyTorch tensor (C, H, W) format
        #rgb_image = torch.tensor(rgb_image).permute(2, 0, 1).float()  # Convert to C, H, W
        if self.transform:
            rgb_image = self.transform(rgb_image)
        
        return rgb_image, label
    
    def split_dataset_based_on_number(samples):
        """Splits the dataset into 3 datasets based on the number in the filename (0, 1, 2)."""
        datasetAxial = []
        datasetCoronial = []
        datasetSagittal = []
        
        # Divide samples into 3 datasets based on the `numberY` (0, 1, 2)
        for img_paths, label, group in samples:
            if group == 0:
                datasetAxial.append((img_paths, label))
            elif group == 1:
                datasetCoronial.append((img_paths, label))
            elif group == 2:
                datasetSagittal.append((img_paths, label))  # 2nd group for datasetC
        
        return datasetAxial, datasetCoronial, datasetSagittal

class ImageTripletDataset_to_three(Dataset):
    def __init__(self, samples, transform=None):
        """
        Args:
            samples (list): List of tuples where each tuple is (img_paths, label).
            transform (callable, optional): Optional transform to be applied on a sample.
        """
        self.class_to_idx = self._create_class_to_idx()
        self.samples = samples
        self.transform = transform

    def _create_class_to_idx(self):
        """Creates a mapping from folder names to integer labels."""
        class_to_idx = {
            'FLAIR': 0,
            'FLAIRCE': 1,
            'OTHER': 2,
            'T1w': 3,
            'T1wCE': 4,
            'T2star': 5,
            'T2w': 6
        }
        return class_to_idx
    
    def __len__(self):
        return len(self.samples)

    def __getitem__(self, idx):
        # Get the paths for the three images and label
        img_paths, label = self.samples[idx]
        
        # Load the three grayscale images
        img0 = Image.open(img_paths[0]).convert('L')  # Convert to grayscale
        img1 = Image.open(img_paths[1]).convert('L')
        img2 = Image.open(img_paths[2]).convert('L')
        
        # Convert to numpy arrays
        img0_array = np.array(img0)
        img1_array = np.array(img1)
        img2_array = np.array(img2)
        
        # Stack into a single multi-channel image (H, W, C)
        rgb_image = np.stack((img0_array, img1_array, img2_array), axis=-1)
        
        # Apply any transformations if specified
        if self.transform:
            rgb_image = self.transform(rgb_image)
        
        return rgb_image, label

class same_anatomical_plane_in_rgb(Dataset):
    def __init__(self, root_dir, transform=None):
        """
        Args:
            root_dir (str): Directory with all the images and subfolders (A, B, C, D, E).
            transform (callable, optional): Optional transform to be applied on a sample.
        """
        self.root_dir = root_dir
        self.transform = transform
        self.class_to_idx = self._create_class_to_idx()
        self.samples = self._load_samples()
        self.targets = [label for _, label in self.samples]

    def _create_class_to_idx(self):
        """Creates a mapping from folder names to integer labels."""
        class_to_idx = {
            'FLAIR': 0,
            'FLAIRCE': 1,
            'OTHER': 2,
            'T1w': 3,
            'T1wCE': 4,
            'T2star': 5,
            'T2w': 6
        }
        return class_to_idx

    def _load_samples(self):
        """Collects all the image paths from the subfolders and associates labels based on folder name."""
        samples = []
        
        # List all the subfolders A, B, C, D, E, etc.
        for folder in os.listdir(self.root_dir):
            folder_path = os.path.join(self.root_dir, folder)
            
            if os.path.isdir(folder_path):  # Check if it is a folder
                # List all files in the folder and sort them
                folder_items =  os.listdir(folder_path)
                main_file_names = set([path.rsplit("_", 2)[0] for path in folder_items])
                for name in main_file_names:
                    if name != "index.html":
                        files_to_connect = sorted([path for path in folder_items if name in path])
                        for i in range(0, len(files_to_connect), 3):
                            img_paths = [os.path.join(folder_path, files_to_connect[i]),
                                        os.path.join(folder_path, files_to_connect[i+1]),
                                        os.path.join(folder_path, files_to_connect[i+2])]

                            label = self.class_to_idx[folder]
                            samples.append((img_paths, label))
        
        return samples

    def __len__(self):
        """Returns the number of samples."""
        return len(self.samples)

    def __getitem__(self, idx):
        """Fetches an item and applies any transformations."""
        # Get the paths for the three images and label
        img_paths, label = self.samples[idx]
        
        # Load the three grayscale images
        img0 = Image.open(img_paths[0]).convert('L')  # Convert to grayscale
        img1 = Image.open(img_paths[1]).convert('L')
        img2 = Image.open(img_paths[2]).convert('L')
        
        # Convert to numpy arrays
        img0_array = np.array(img0)
        img1_array = np.array(img1)
        img2_array = np.array(img2)
        
        # Stack into an RGB image
        rgb_image = np.stack((img0_array, img1_array, img2_array), axis=-1)
        
        # Convert numpy array to PyTorch tensor (C, H, W) format
        #rgb_image = torch.tensor(rgb_image).permute(2, 0, 1).float()  # Convert to C, H, W
        if self.transform:
            rgb_image = self.transform(rgb_image)
        
        return rgb_image, label
