import torch
import numpy as np
import torch.nn.functional as F
from collections import defaultdict

from parameters import DEVICE

def predict(model, data_loader):
    model.eval()

    # save the predictions in this list
    y_pred = []

    # no gradient needed
    with torch.no_grad():
        gt_labels_conf = []
        # go over each batch in the loader. We can ignore the targets here
        for batch, gt_label in data_loader:

            # Move batch to the GPU
            batch = batch.to(DEVICE)

            # predict probabilities of each class
            predictions = model(batch)

            # apply a softmax to the predictions
            predictions = F.softmax(predictions, dim=1)

            # move to the cpu and convert to numpy
            predictions = predictions.cpu().numpy()

            # save
            y_pred.append(predictions)
            # Get batch indices
            batch_indices = np.arange(len(gt_label))
            # Extract ground truth confidences
            gt_confidences = predictions[batch_indices, gt_label.numpy()]
            gt_labels_conf.extend(gt_confidences)
            #gt_labels_conf.append(predictions[gt_label])
    # stack predictions into a (num_samples, 10) array
    y_pred = np.vstack(y_pred)
    return y_pred, gt_labels_conf

def averaging_prediction(y_pred, test_set, test_image_paths):
    gt_y = np.array(test_set.targets) # get the true labels and convert to numpy
    grouped_predictions = defaultdict(list)
    grouped_gt_predictions = defaultdict(list)
    ground_truth_dict = {}
    
    for i, image_path in enumerate(test_image_paths):
        # Strip last 5 characters (this should group the same items)
        if isinstance(image_path, list):
            base_name = image_path[0][:-10]
        else:
            base_name = image_path[:-10]
        # Append the prediction probabilities for the current image to the corresponding group
        grouped_predictions[base_name].append(y_pred[i])
        grouped_gt_predictions[base_name].append(y_pred[i][gt_y[i]])

        if base_name not in ground_truth_dict:
            ground_truth_dict[base_name] = gt_y[i]
        elif ground_truth_dict[base_name] != gt_y[i]:
            print(f"WARNING: The ground truth label differs from the same image but from different views")

    final_predictions = {}

    for base_name, preds in grouped_predictions.items():
        preds_array = np.array(preds)
        
        combined_probs = np.sum(preds_array, axis=0) / 3
        
        final_class = np.argmax(combined_probs)
        final_predictions[base_name] = final_class

    return grouped_predictions, ground_truth_dict, final_predictions, grouped_gt_predictions

def predict_ENSEMBLE_with_confidence(model1, model2, model3, data_loader1, data_loader2, data_loader3, device):
    # Set models to evaluation mode
    model1.eval()
    model2.eval()
    model3.eval()

    # To store final predictions, true labels, and confidence scores
    all_majority_preds = []
    true_labels = []
    confidence_scores_majority = []  # Confidence of the majority-voted prediction
    confidence_scores_true = []      # Confidence of the ground truth label

    # No gradient calculation needed
    with torch.no_grad():
        # Iterate over each batch from the three data loaders in sync
        for (batch1, targets1), (batch2, _), (batch3, _) in zip(data_loader1, data_loader2, data_loader3):
            # Extend true labels from the first loader
            true_labels.extend(targets1.cpu().numpy())

            # Move each batch to the device
            batch1, batch2, batch3 = batch1.to(device), batch2.to(device), batch3.to(device)

            # Get predictions from each model and apply softmax to get probabilities
            probs1 = F.softmax(model1(batch1), dim=1).cpu().numpy()
            probs2 = F.softmax(model2(batch2), dim=1).cpu().numpy()
            probs3 = F.softmax(model3(batch3), dim=1).cpu().numpy()

            # Stack probabilities to form an array of shape (batch_size, 3, num_classes)
            stacked_probs = np.stack([probs1, probs2, probs3], axis=1)  # (batch_size, 3, num_classes)

            # Compute the average probabilities across models for each class
            avg_probs = np.mean(stacked_probs, axis=1)  # (batch_size, num_classes)

            # Get the predicted class (majority) and confidence for each sample
            majority_preds = avg_probs.argmax(axis=1)
            confidence_scores_majority.extend(avg_probs.max(axis=1))

            # Get confidence scores for the true labels
            true_confidences = avg_probs[np.arange(len(targets1)), targets1.cpu().numpy()]
            confidence_scores_true.extend(true_confidences)

            # Collect final majority-voted predictions
            all_majority_preds.extend(majority_preds)

    # Convert lists to numpy arrays for easy analysis if needed
    all_majority_preds = np.array(all_majority_preds)
    true_labels = np.array(true_labels)
    confidence_scores_majority = np.array(confidence_scores_majority)
    confidence_scores_true = np.array(confidence_scores_true)

    return all_majority_preds, true_labels, confidence_scores_majority, confidence_scores_true

