import torch
import numpy as np

def train_for_epoch_with_scaler(model, train_loader, optimizer, criterion, scaler, device):
    # set model to train
    model.train()
    
    train_losses = []
    train_accuracies = []
    counter = 0

    for batch, target in train_loader:

        # data to GPU
        batch = batch.to(device)
        target = target.to(device)

        # reset optimizer
        optimizer.zero_grad()

        # forward pass
        predictions = model(batch)

        # calculate accuracy
        accuracy = (torch.argmax(predictions, dim=1) == target).sum().item() / target.size(0)
        
        # calculate loss
        loss = criterion(predictions, target)

        # backward pass
        scaler.scale(loss).backward()

        # parameter update
        scaler.step(optimizer)
        scaler.update()

        # track loss
        train_losses.append(float(loss.item()))
        train_accuracies.append(accuracy)

        counter += 1
        if counter % 20 == 0:
          print('[{}/{} ({:.0f}%)]\tLoss: {:.6f}'.format(
                int(counter * len(batch)), len(train_loader.dataset),
                100. * counter / len(train_loader), loss.item()))

    train_loss = np.mean(np.array(train_losses))
    train_accuracy = np.mean(np.array(train_accuracies))
    
    print('\nTrain: Average loss: {:.4f}, Accuracy: {:.4f}\n'.format(
        train_loss, train_accuracy))
    
    return train_loss, train_accuracy

def validate(model, val_loader, criterion, device):
    model.eval()
    
    val_losses = []
    y_true, y_pred = [], []

    with torch.no_grad():
        for batch, target in val_loader:

            # move data to the device
            batch = batch.to(device)
            target = target.to(device)

            with torch.autocast(device_type=device, dtype=torch.float16):
              # make predictions
              predictions = model(batch)

              # calculate loss
              loss = criterion(predictions, target)

            # track losses and predictions
            val_losses.append(float(loss.item()))
            y_true.extend(target.cpu().numpy())
            y_pred.extend(predictions.argmax(dim=1).cpu().numpy())

    y_true = np.array(y_true)
    y_pred = np.array(y_pred)
    val_losses = np.array(val_losses)

    # calculate validation accuracy from y_true and y_pred
    val_accuracy = np.mean(y_true == y_pred)

    # calculate the mean validation loss
    val_loss = np.mean(val_losses)

    print('Validation: Average loss: {:.4f}, Accuracy: {:.4f}\n'.format(
        val_loss, val_accuracy))

    return val_loss, val_accuracy

def train_with_scaler(model, train_loader, val_loader, optimizer, criterion, epochs, scaler, device, checkpoints_foler = None, first_epoch=1):
    train_losses, val_losses = [],  []
    train_accuracies, val_accuracies = [], []
    max_val_acc = 0
    best_epoch = 0

    for epoch in range(first_epoch, epochs+first_epoch):

        print('Train Epoch: {}'.format(epoch))

        # train
        train_loss, train_acc = train_for_epoch_with_scaler(model, train_loader, optimizer, criterion, scaler, device)

        # validation
        valid_loss, valid_acc = validate(model, val_loader, criterion, device)

        train_losses.append(train_loss)
        val_losses.append(valid_loss)
        train_accuracies.append(train_acc)
        val_accuracies.append(valid_acc)

        # save checkpoint
        if checkpoints_foler != None and max_val_acc < valid_acc:
          max_val_acc = valid_acc
          best_epoch = epoch
          torch.save(model, checkpoints_foler+f'/avp_{epoch:03d}.pkl')

    return train_losses, val_losses, train_accuracies, val_accuracies, best_epoch