# Descriptrion!!!!

import torch
import torchvision.transforms as transforms
import torch.nn.functional as F
from torchvision.datasets import ImageFolder
from torch.utils.data import DataLoader
from torchvision.transforms import v2
torch.backends.cudnn.deterministic = True
torch.backends.cudnn.benchmark = False

import numpy as np
import pandas as pd
from sklearn.metrics import confusion_matrix
from itertools import product

import random
random.seed(42)
torch.manual_seed(42)
np.random.seed(42)

# define custom resample class to change image resolution without rescaling
class RandomResample:
    def __init__(self, scale_factor):
        self.scale_factor = random.uniform(0,scale_factor)
    
    def __call__(self, img):
        # Downsample
        width, height = img.size
        downscaled_size = (int(width / self.scale_factor), int(height / self.scale_factor))
        
        # Downsample the image
        img_downsampled = img.resize(downscaled_size)
        
        # Upsample back to the original size
        img_upsampled = img_downsampled.resize((width, height))
        
        return img_upsampled
    
# Define a function to get the transformations based on the configuration
def get_transform(use_hflip, use_vflip, rotation, contrast, resample):
    transform_list = []
    # resize
    transform_list.append(transforms.Resize(224))
    # horizontal flip
    if use_hflip:
        transform_list.append(transforms.RandomHorizontalFlip())
    # vertical flip
    if use_vflip:
        transform_list.append(transforms.RandomVerticalFlip())
    # rotation
    transform_list.append(transforms.RandomRotation(degrees=rotation, expand=True)) # expand=True: esnure that the whole image is represented on the rotated image
    # contrast
    if contrast > 0:
        transform_list.append(v2.ColorJitter(contrast=contrast))
    # resample
    if resample != 1:
        transform_list.append(RandomResample(scale_factor=resample))
    transform_list.append(transforms.Resize(224))
    transform_list.append(transforms.ToTensor())
    return transforms.Compose(transform_list)


def predict(model, data_loader, device):
    model.eval()

    # save the predictions in this list
    y_pred = []

    # no gradient needed
    with torch.no_grad():

        # go over each batch in the loader. We can ignore the targets here
        for batch, _ in data_loader:

            # Move batch to the GPU
            batch = batch.to(device)

            # predict probabilities of each class
            predictions = model(batch)

            # apply a softmax to the predictions
            predictions = F.softmax(predictions, dim=1)

            # move to the cpu and convert to numpy
            predictions = predictions.cpu().numpy()

            # save
            y_pred.append(predictions)

    # stack predictions into a (num_samples, 10) array
    y_pred = np.vstack(y_pred)
    return y_pred


# MAIN
if torch.cuda.is_available():
    device = 'cuda'
    print('cuda is available')
else:
    device = 'cpu'
    print('cuda is not available, change to cpu')
    

model_path = '/net/cremi/bformanek/TRDP_II/local_models/'
#model_name = 'transfer_checkpoints_resnet18_adam_amp_criterion_balanced_avp_025.pkl' # original resnet18
#model_name = 'transfer_checkpoints_resnet18_flips_avp_022.pkl' # augmented with flips
#model_name = 'transfer_checkpoints_resnet18_flips_90_avp_013.pkl' # augmented with flips and rotation upto 90 degrees
#model_name = 'transfer_checkpoints_resnet18_flips_180_avp_022.pkl' # augmented with flips and rotation upto 180 degrees
#model_name = 'transfer_checkpoints_resnet18_flips_360_avp_022.pkl' # augmented with flips and rotation upto 360 degrees
model_name = 'transfer_checkpoints_resnet18_flips_90_resample_avp_016.pkl' # augmented with flips and rotation upto 90 degrees plus resample upto half resolution
model = torch.load(model_path + model_name, map_location=torch.device('cpu'), weights_only=False)
model.to(device)

DATA_PATH = '/net/travail/bformanek/MRI_dataset'
TEST_FOLDER = DATA_PATH + '/test'

# Set up your transformations with a list of options
use_random_hflip = [False, True]
use_random_vflip = [False, True]
random_rotation = list(range(0, 180, 10))
#random_contrast = [0, 0.1, 0.2, 0.3, 0.4, 0.5]
random_contrast = [0]
use_resample = [1, 2, 4]

BATCH_SIZE = 64
WORKERS = 8

# Initialize list to store results
results = []
iterations = 0
y_true_set = False # flag to calculate y_true only once
for hflip, vflip, rotation, contrast, resample in product(use_random_hflip, use_random_vflip, random_rotation, random_contrast, use_resample):
    # skip iterations at the beginning if needed
    if iterations < 180:
      iterations += 1
      continue
    
    transform = get_transform(hflip, vflip, rotation, contrast, resample)

    test_set = ImageFolder(TEST_FOLDER, transform = transform)
    test_loader = DataLoader(test_set, batch_size = BATCH_SIZE,  shuffle = False, num_workers=WORKERS)
    
    # compute predictions on the test set
    y_pred = predict(model, test_loader, device)
    
    # find the argmax of each of the predictions
    y_pred = y_pred.argmax(axis=1)

    # get the true labels and convert to numpy, but only if it is not already set     
    if not y_true_set:
        y_true = np.array(test_set.targets)
        y_true_set = True
    
    # balanced accuracy    
    conf_matrix = confusion_matrix(y_true, y_pred)
    TP = conf_matrix.diagonal()
    P = conf_matrix.sum(axis=1)
    balanced_accuracy = sum(TP / P) / len(P)
    
    results.append({
        "use_random_hflip": hflip,
        "use_random_vflip": vflip,
        "random_rotation": rotation,
        "random_contrast": contrast,
        "resample": resample,
        "balanced_accuracy": balanced_accuracy
    })
    
    print(f"use_random_hflip - {hflip}, use_random_vflip - {vflip}, random_rotation - {rotation}, random_contrast - {contrast}, resample - {resample}, balanced_accuracy - {balanced_accuracy}")
    
    iterations += 1
    
    if iterations % 36 == 0:
        # Save results to a DataFrame and export to CSV for easy analysis
        df_results = pd.DataFrame(results)
        df_results.to_csv(f"transform_evaluation_results_resnet18_flips_90_resample_{iterations}.csv", index=False) 
        print(f"save at iteration {iterations}")       

    del test_set, test_loader, y_pred
    
# Save results to a DataFrame and export to CSV for easy analysis
df_results = pd.DataFrame(results)
df_results.to_csv("transform_evaluation_results_resnet18_flips_90_resample.csv", index=False)