# Infernece time measurement:
# Measure inference times for a set of models in a folder (local_model_folder) with random sample images from the datast (test_image_folder)
# Needed: 
#   - models loaded into local folder (use copy_best_models_to_local.py)
#   - define local and server folders, and the number of measurements
# Script:
#   0) iterate through the number of measurements
#   1) copy a random image from the dataset to local (for moer reliable masurement results)
#   2) evaluate cpu and gpu inference times and gpu memory for every model on the sample image
#       - on CPU do not use autocast, as the CREMI CPU-s do not support this option
#       - on GPU use autocast with float16, and calculate memory usage
#   3) put results to command line
#   4) remove sample image from local
# Results:
#   Log rasults can be found in ./time_measurement folder

import torch
import torchvision.transforms as transforms
from torchvision.datasets import ImageFolder
from torch.utils.data import DataLoader
import time
import random
import os
from PIL import Image
import shutil

# Get a random image path from given folder
def get_random_image_path(folder):
    # Get all subfolders and files
    all_images = []
    for subdir, _, files in os.walk(folder):
        for file in files:
            if file.endswith(('jpg', 'png', 'jpeg')):
                all_images.append(os.path.join(subdir, file))
    
    # Choose a random image path
    random_image_path = random.choice(all_images)
    return random_image_path

def copy_image_to_local(remote_image_folder, local_images_folder):
    # Randomly select an image path, in other case it might be given by the user
    image_path = get_random_image_path(remote_image_folder)
    print(f"Selected image path: {image_path}")
    
    # copy image to local 
    local_image_path = local_images_folder + os.path.basename(image_path)
    shutil.copyfile(image_path, local_image_path)
    print(f"Copied to local image path: {local_image_path}")
    
    return local_image_path
    
def inference_time_measurement(model_path, image_path, transform, device = 'cuda', num_wokres = 8):
    # check cuda
    if device == 'cuda':
        if torch.cuda.is_available():
            print('cuda is available')
        else:
            print('cuda is not available, change to cpu')
            device = 'cpu'
    print("Inference time masurement using " + device)
    
    if device == 'cuda':
        # Measure memory before prediction
        torch.cuda.synchronize()
        start_memory = torch.cuda.memory_allocated(device)
        
    # 1) load and transform image
    start_time = time.time()
    load_image_time_start = start_time

    # Load the image and apply transformations
    image = Image.open(image_path).convert('RGB')
    image = transform(image).unsqueeze(0)  # Add batch dimension

    load_image_time_end = time.time()
    load_image_time = (load_image_time_end - load_image_time_start) * 1000
    print(f'Image loaded: {load_image_time:.2f} ms')

    # 2) Measure Model Loading Time
    # Get the model name
    # print("Loading model")

    load_model_time_start = time.time() # start timer
    model = torch.load(model_path, map_location=torch.device('cpu'), weights_only=False)
    model.eval() # Set model to evaluation mode
    model.to(device)
    load_model_time_end = time.time()

    load_model_time = (load_model_time_end - load_model_time_start) * 1000 # elapsed time in milliseconds
    print(f'Model loaded: {load_model_time:.2f} ms')

    # 3) Predict on the single image
    image = image.to(device)
    prediction_time_start = time.time()
    with torch.no_grad():
        if device == 'cuda':
            with torch.autocast(device_type=device, dtype=torch.bfloat16):
                # Perform prediction
                output = model(image)
        else:
            output = model(image)
            
    end_time = time.time()
    prediction_time = (end_time - prediction_time_start) * 1000
    print(f'Prediction time for one image: {prediction_time:.2f} ms')

    inference_time = (end_time - start_time) * 1000
    print(f'Compete inference time for one image: {inference_time:.2f} ms')

    if device == 'cuda':
        # Measure memory after prediction
        end_memory = torch.cuda.memory_allocated(device)
        
        print(f'GPU memory usage: {(end_memory - start_memory)/(1024**2)} MB')
        
        
# Main function
number_of_measurements = 4

# Define the transformation for the test data
transform = transforms.Compose([
    transforms.Resize(224),
    transforms.ToTensor()
])

# Define folders 
test_image_folder = '/net/travail/bformanek/MRI_dataset/test'
local_images_folder = '/net/cremi/bformanek/TRDP_II/local_images/'
local_model_folder = '/net/cremi/bformanek/TRDP_II/local_models/'

"""
predefined_images_list = [
    'T1w/4RTNI_5_S_5006_MR_Sag_IR-SPGR__br_raw_20140821092148809_194_S228381_I440459_2529.7502_0_093.png',
    'T1w/4RTNI_5_S_5006_MR_Sag_IR-SPGR__br_raw_20140821092148809_194_S228381_I440459_2529.7502_0_093.png',
    'T1w/ISYB_sub-0103_T1w_3526.7072_2_121.png',
    'T1w/ADNI_035_S_0048_MR_MPR-R__GradWarp__B1_Correction__N3_Br_20070319172419398_S10258_I45187_617.3637_0_091.png'
]"""

for i in range(number_of_measurements):
    print(f"MEASUREMENT - {i}" )
    # get image to local folder
    local_image_path = copy_image_to_local(test_image_folder, local_images_folder) # random image
    
    # local_image_path = os.path.join(local_images_folder, os.path.basename(predefined_images_list[i])) # predefined image
    # shutil.copyfile(os.path.join(test_image_folder, predefined_images_list[i]), local_image_path) # predefined image
    
    for model_name in os.listdir(local_model_folder) :
        print("MODEL: ")
        print(model_name)
        
        for device in ['cpu', 'cuda']:
            print("DEVICE: " + device)
            
            if device == 'cuda':
                if not torch.cuda.is_available():
                    print('cuda is not available')
                    break

            model_path = local_model_folder + model_name
            
            inference_time_measurement(model_path = model_path, 
                                    image_path = local_image_path,
                                    transform = transform,
                                    device = device,
                                    num_wokres = 8)
        print("-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ")
    print("-- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- -- ")
    os.remove(local_image_path)